@@ -2,7 +2,7 @@
 package MRTG_lib;
 
 ###################################################################
-# MRTG 2.10.5  Support library MRTG_lib.pm
+# MRTG 2.10.8  Support library MRTG_lib.pm
 ###################################################################
 # Created by Tobias Oetiker <oetiker@ee.ethz.ch>
 #            and Dave Rand <dlr@bungi.com>
@@ -32,7 +32,7 @@ BEGIN {
         $PS = ':';
     } elsif ( $^O =~ /^os2$/i ) {
 	$OS = 'OS2';
-	$SL = '\\';
+	$SL = '/';
 	$PS = ';';
     }  else {
         $OS = 'UNIX';
@@ -971,7 +971,7 @@ sub expistr ($) {
 
 sub create_pid ($) {
     my $pidfile = shift;
-    return if ($OS eq 'NT' || $OS eq 'OS2');
+    return if ($OS eq 'NT' );
     return if -e $pidfile;
     if ( open(PIDFILE,">$pidfile")) {
          close PIDFILE;
@@ -982,6 +982,7 @@ sub create_pid ($) {
 
 sub demonize_me ($) {
     my $pidfile = shift;
+    my $cfgfile = shift;
     print "Daemonizing MRTG ...\n";
     if ( $OS eq 'NT' ) {
         print "Do Not close this window. Or MRTG will die\n";
@@ -992,12 +993,24 @@ sub demonize_me ($) {
 #            $CONSOLE->Free();
 #            $CONSOLE->Alloc();
 #            $CONSOLE->Mode()
+    }
+    elsif( $OS eq 'OS2')
+    {
+     require OS2::Process;
+     if (my_type() eq 'VIO'){
+        $main::Cleanfile3 = $pidfile;
+
+        print "MRTG detached. PID=".system(P_DETACH,$^X." ".$0." ".$cfgfile);
+        exit;
+     }
     } else {
            # Check out if there is another mrtg running before forking
-           if (defined $pidfile && open(READPID, "<$pidfile")) {
-               chomp(my $input = <READPID>);    # read process id in pidfile
-               if ($input && kill 0 => $input) {# oops - the pid actually exists
-                   die "ERROR: I Quit! Another copy of mrtg seems to be running. Check $pidfile\n";
+           if (defined $pidfile && open(READPID, "<$pidfile")){
+               if (not eof READPID) {
+                   chomp(my $input = <READPID>);    # read process id in pidfile
+                   if ($input && kill 0 => $input) {# oops - the pid actually exists
+                        die "ERROR: I Quit! Another copy of mrtg seems to be running. Check $pidfile\n";
+                   }
                }
                close READPID;
            }
@@ -1016,7 +1029,7 @@ sub demonize_me ($) {
                    }
               }
               require 'POSIX.pm';
-              &POSIX::setsid or die "Can't start a new session: $!";
+              POSIX::setsid() or die "Can't start a new session: $!";
               open STDOUT,'>/dev/null' or die "ERROR: Redirecting STDOUT to /dev/null: $!";
               open STDIN, '</dev/null' or die "ERROR: Redirecting STDIN from /dev/null: $!";
       }
